@@ -2,12 +2,15 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at https://mozilla.org/MPL/2.0/. */
 
+use crate::generated::{GeneratedProperty, GeneratedType, SerdeOptions};
 use crate::parser::{
     parse_from_file, AllOf, AnyOf, DataType, Object, ObjectProperty, OneOf, PrimitiveType, Ref,
     Root,
 };
 use crate::resolver::{ResolveResult, Resolver};
 use crate::sanitizer::{sanitize_property_name, sanitize_struct_name};
+use proc_macro2::TokenStream;
+use quote::quote;
 use std::cmp::Ordering;
 use std::collections::HashMap;
 use std::path::Path;
@@ -31,66 +34,47 @@ impl<T: Eq> PartialOrd for EntryWithPosition<T> {
     }
 }
 
-#[derive(Eq, PartialEq, Debug)]
-struct Type {
-    src: String,
-    name: String,
-    properties: Vec<Property>,
-}
-
-impl Type {
-    pub fn serialize(&self) -> String {
-        let mut result = String::from("");
-
-        result.push_str(&format!("// from {}\n", self.src));
-        result.push_str("#[derive(Clone, PartialEq, Debug, Deserialize, Serialize)]\n");
-        result.push_str(&format!("struct {} {{\n", self.name));
-
-        let properties: Vec<String> = self.properties.iter().map(|x| x.serialize()).collect();
-        result.push_str(&properties.join("\n"));
-
-        result.push_str("}");
-
-        result
-    }
-}
-
-#[derive(Eq, PartialEq, Debug)]
-struct Property {
-    name: String,
-    property_type: String,
-    serde_options: SerdeOptions,
-}
-
-impl Property {
-    pub fn serialize(&self) -> String {
-        let mut result = String::from("");
-
-        match &self.serde_options.rename {
-            Some(rename) => {
-                result.push_str(&format!("    #[serde(rename = \"{}\")]\n", rename));
-            }
-            None => {}
-        }
-
-        result.push_str(&format!("    pub {}: {},\n", self.name, self.property_type));
-
-        result
-    }
-}
-
-#[derive(Eq, PartialEq, Debug)]
-struct SerdeOptions {
-    rename: Option<String>,
-}
-
 pub struct Generator {
     resolver: Resolver,
-    types: HashMap<String, EntryWithPosition<Type>>,
+    types: HashMap<String, EntryWithPosition<GeneratedType>>,
     next_position: u64,
     known_type_names: HashMap<String, String>,
 }
 
+impl Into<Vec<GeneratedType>> for Generator {
+    fn into(self) -> Vec<GeneratedType> {
+        let mut types: Vec<EntryWithPosition<GeneratedType>> =
+            self.types.into_iter().map(|(_, value)| value).collect();
+
+        types.sort();
+
+        types
+            .into_iter()
+            .map(
+                |EntryWithPosition {
+                     payload,
+                     position: _,
+                 }| payload,
+            )
+            .collect()
+    }
+}
+
+impl Into<TokenStream> for Generator {
+    fn into(self) -> TokenStream {
+        let types: Vec<GeneratedType> = self.into();
+
+        let tokens: Vec<TokenStream> = types.into_iter().map(|x| x.into()).collect();
+
+        quote! {
+            use serde_derive::{Serialize, Deserialize};
+            use serde_json::Value;
+            use std::collections::BTreeMap;
+            #(#tokens)*
+        }
+    }
+}
+
 impl Generator {
     pub fn new() -> Self {
         Generator {
@@ -163,7 +147,7 @@ impl Generator {
                         ));
                     }
 
-                    let new_type = Type {
+                    let new_type = GeneratedType {
                         src: src.clone(),
                         name: name.clone(),
                         properties: new_properties,
@@ -210,7 +194,7 @@ impl Generator {
             data_type,
         }: &ObjectProperty,
         visited_objects: Vec<String>,
-    ) -> Property {
+    ) -> GeneratedProperty {
         let property_name = sanitize_property_name(name.clone());
 
         let rename = if name == &property_name {
@@ -219,7 +203,7 @@ impl Generator {
             Some(name.clone())
         };
 
-        Property {
+        GeneratedProperty {
             name: property_name,
             property_type: self.add_type(
                 base_path,
@@ -328,29 +312,17 @@ impl Generator {
             DataType::Any => String::from("Value"),
         }
     }
-
-    pub fn serialize(&self) -> String {
-        let mut result = String::from("");
-
-        result.push_str("use serde_json::Value;\n\n");
-        result.push_str("use std::collections::BTreeMap;\n\n");
-
-        let mut types: Vec<&EntryWithPosition<Type>> = self.types.values().collect();
-        types.sort();
-        let types: Vec<String> = types.into_iter().map(|x| x.payload.serialize()).collect();
-        result.push_str(&types.join("\n\n"));
-        result.push_str("\n");
-
-        result
-    }
 }
 
 #[cfg(test)]
 mod generator_tests {
-    use crate::generator::{EntryWithPosition, Generator, Property, SerdeOptions, Type};
+    use crate::generator::{
+        EntryWithPosition, GeneratedProperty, GeneratedType, Generator, SerdeOptions,
+    };
     use crate::parser::{
         AllOf, AnyOf, DataType, Object, ObjectProperty, OneOf, PrimitiveType, Ref, Root,
     };
+    use proc_macro2::TokenStream;
     use std::collections::HashMap;
     use std::path::Path;
     use std::rc::Rc;
@@ -393,103 +365,23 @@ mod generator_tests {
         );
     }
 
-    #[test]
-    fn should_serialize_type_with_derive() {
-        let t = Type {
-            src: String::from(""),
-            name: String::from(""),
-            properties: Vec::new(),
-        };
-
-        assert_eq!(
-            t.serialize()
-                .contains("#[derive(Clone, PartialEq, Debug, Deserialize, Serialize)]"),
-            true
-        )
-    }
-
-    #[test]
-    fn should_serialize_type_with_src() {
-        let t = Type {
-            src: String::from("foo"),
-            name: String::from(""),
-            properties: Vec::new(),
-        };
-
-        assert_eq!(t.serialize().contains("from foo"), true)
-    }
-
-    #[test]
-    fn should_serialize_type_with_struct() {
-        let t = Type {
-            src: String::from(""),
-            name: String::from("Foo"),
-            properties: Vec::new(),
-        };
-
-        assert_eq!(t.serialize().contains("struct Foo"), true)
-    }
-
-    #[test]
-    fn should_serialize_type_with_properties() {
-        let t = Type {
-            src: String::from(""),
-            name: String::from(""),
-            properties: vec![Property {
-                name: String::from("foo"),
-                property_type: String::from("Obsidian"),
-                serde_options: SerdeOptions { rename: None },
-            }],
-        };
-
-        assert_eq!(t.serialize().contains("foo: Obsidian"), true)
-    }
-
-    #[test]
-    fn should_serialize_property_with_name() {
-        let property = Property {
-            name: String::from("foo"),
-            property_type: String::from("Obsidian"),
-            serde_options: SerdeOptions { rename: None },
-        };
-
-        assert_eq!(property.serialize().contains("pub foo: Obsidian"), true)
-    }
-
-    #[test]
-    fn should_serialize_property_with_rename() {
-        let property = Property {
-            name: String::from(""),
-            property_type: String::from(""),
-            serde_options: SerdeOptions {
-                rename: Some(String::from("foo")),
-            },
-        };
-
-        assert_eq!(
-            property.serialize().contains("#[serde(rename = \"foo\")]"),
-            true
-        )
-    }
-
     #[test]
     fn should_serialize_with_serde_json_import() {
         let generator = Generator::new();
+        let tokens: TokenStream = generator.into();
 
-        assert_eq!(
-            generator.serialize().contains("use serde_json::Value"),
-            true
-        )
+        assert_eq!(tokens.to_string().contains("use serde_json :: Value"), true)
     }
 
     #[test]
     fn should_serialize_with_btree_import() {
         let generator = Generator::new();
+        let tokens: TokenStream = generator.into();
 
         assert_eq!(
-            generator
-                .serialize()
-                .contains("use std::collections::BTreeMap"),
+            tokens
+                .to_string()
+                .contains("use std :: collections :: BTreeMap"),
             true
         )
     }
@@ -506,10 +398,10 @@ mod generator_tests {
             generator.types.get("correct src"),
             Some(&EntryWithPosition {
                 position: 0,
-                payload: Type {
+                payload: GeneratedType {
                     src: String::from("correct src"),
                     name: String::from("AwesomeFoo"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("awesome_property"),
                         property_type: String::from("Value"),
                         serde_options: SerdeOptions {
@@ -644,14 +536,18 @@ mod generator_tests {
     }
 
     fn object_with_property() -> Object {
+        object_with_custom_property(ObjectProperty {
+            name: String::from("awesome property"),
+            required: false,
+            data_type: Rc::new(DataType::Any),
+        })
+    }
+
+    fn object_with_custom_property(property: ObjectProperty) -> Object {
         Object {
             src: String::from("wrong src"),
             name: String::from("awesome foo"),
-            properties: vec![ObjectProperty {
-                name: String::from("awesome property"),
-                required: false,
-                data_type: Rc::new(DataType::Any),
-            }],
+            properties: vec![property],
         }
     }
 
@@ -862,7 +758,7 @@ mod generator_tests {
         let mut generator = Generator::new();
         generator.add_file(Path::new(file));
 
-        let mut types: Vec<EntryWithPosition<Type>> = generator
+        let mut types: Vec<EntryWithPosition<GeneratedType>> = generator
             .types
             .into_iter()
             .map(|(_, value)| value)
@@ -870,33 +766,33 @@ mod generator_tests {
 
         types.sort();
 
-        let types: Vec<Type> = types.into_iter().map(|x| x.payload).collect();
+        let types: Vec<GeneratedType> = types.into_iter().map(|x| x.payload).collect();
 
         assert_eq!(
             types,
             vec![
-                Type {
+                GeneratedType {
                     src: String::from("src/examples/generator/loop1.schema.json"),
                     name: String::from("Loop"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("a"),
                         serde_options: SerdeOptions { rename: None },
                         property_type: String::from("Option<B>")
                     }]
                 },
-                Type {
+                GeneratedType {
                     src: String::from("src/examples/generator/loop1.schema.json#/definitions/b"),
                     name: String::from("B"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("c"),
                         serde_options: SerdeOptions { rename: None },
                         property_type: String::from("Option<C>")
                     }]
                 },
-                Type {
+                GeneratedType {
                     src: String::from("src/examples/generator/loop2.schema.json#/definitions/c"),
                     name: String::from("C"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("b"),
                         serde_options: SerdeOptions { rename: None },
                         property_type: String::from("Option<Box<B>>")
@@ -913,7 +809,7 @@ mod generator_tests {
         let mut generator = Generator::new();
         generator.add_file(Path::new(file));
 
-        let mut types: Vec<EntryWithPosition<Type>> = generator
+        let mut types: Vec<EntryWithPosition<GeneratedType>> = generator
             .types
             .into_iter()
             .map(|(_, value)| value)
@@ -921,31 +817,31 @@ mod generator_tests {
 
         types.sort();
 
-        let types: Vec<Type> = types.into_iter().map(|x| x.payload).collect();
+        let types: Vec<GeneratedType> = types.into_iter().map(|x| x.payload).collect();
 
         assert_eq!(
             types,
             vec![
-                Type {
+                GeneratedType {
                     src: String::from(file),
                     name: String::from("Twice"),
                     properties: vec![
-                        Property {
+                        GeneratedProperty {
                             name: String::from("a"),
                             serde_options: SerdeOptions { rename: None },
                             property_type: String::from("Option<C>")
                         },
-                        Property {
+                        GeneratedProperty {
                             name: String::from("b"),
                             serde_options: SerdeOptions { rename: None },
                             property_type: String::from("Option<C>")
                         }
                     ]
                 },
-                Type {
+                GeneratedType {
                     src: String::from(format!("{}#/definitions/c", file)),
                     name: String::from("C"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("foo"),
                         serde_options: SerdeOptions { rename: None },
                         property_type: String::from("Value")
@@ -962,7 +858,7 @@ mod generator_tests {
         let mut generator = Generator::new();
         generator.add_file(Path::new(file));
 
-        let mut types: Vec<EntryWithPosition<Type>> = generator
+        let mut types: Vec<EntryWithPosition<GeneratedType>> = generator
             .types
             .into_iter()
             .map(|(_, value)| value)
@@ -970,54 +866,54 @@ mod generator_tests {
 
         types.sort();
 
-        let types: Vec<Type> = types.into_iter().map(|x| x.payload).collect();
+        let types: Vec<GeneratedType> = types.into_iter().map(|x| x.payload).collect();
 
         assert_eq!(
             types,
             vec![
-                Type {
+                GeneratedType {
                     src: String::from(file),
                     name: String::from("Collision"),
                     properties: vec![
-                        Property {
+                        GeneratedProperty {
                             name: String::from("a"),
                             serde_options: SerdeOptions { rename: None },
                             property_type: String::from("Option<A>")
                         },
-                        Property {
+                        GeneratedProperty {
                             name: String::from("b"),
                             serde_options: SerdeOptions { rename: None },
                             property_type: String::from("Option<A1>")
                         },
-                        Property {
+                        GeneratedProperty {
                             name: String::from("c"),
                             serde_options: SerdeOptions { rename: None },
                             property_type: String::from("Option<A2>")
                         }
                     ]
                 },
-                Type {
+                GeneratedType {
                     src: String::from(format!("{}/properties/a", file)),
                     name: String::from("A"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("foo"),
                         serde_options: SerdeOptions { rename: None },
                         property_type: String::from("Value")
                     }]
                 },
-                Type {
+                GeneratedType {
                     src: String::from(format!("{}/properties/b", file)),
                     name: String::from("A1"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("foo"),
                         serde_options: SerdeOptions { rename: None },
                         property_type: String::from("Value")
                     }]
                 },
-                Type {
+                GeneratedType {
                     src: String::from(format!("{}/properties/c", file)),
                     name: String::from("A2"),
-                    properties: vec![Property {
+                    properties: vec![GeneratedProperty {
                         name: String::from("foo"),
                         serde_options: SerdeOptions { rename: None },
                         property_type: String::from("Value")
@@ -1027,6 +923,51 @@ mod generator_tests {
         );
     }
 
+    #[test]
+    fn should_convert_into_a_sorted_type_list() {
+        let mut generator = Generator::new();
+
+        add_type(
+            &mut generator,
+            DataType::Object(object_with_custom_property(ObjectProperty {
+                name: String::from("first property"),
+                required: false,
+                data_type: Rc::new(DataType::Object(object_with_property())),
+            })),
+            true,
+        );
+
+        let types: Vec<GeneratedType> = generator.into();
+
+        assert_eq!(
+            types,
+            vec![
+                GeneratedType {
+                    src: String::from(""),
+                    name: String::from("AwesomeFoo"),
+                    properties: vec![GeneratedProperty {
+                        name: String::from("first_property"),
+                        serde_options: SerdeOptions {
+                            rename: Some(String::from("first property"))
+                        },
+                        property_type: String::from("Option<AwesomeFoo1>")
+                    }]
+                },
+                GeneratedType {
+                    src: String::from("wrong src"),
+                    name: String::from("AwesomeFoo1"),
+                    properties: vec![GeneratedProperty {
+                        name: String::from("awesome_property"),
+                        serde_options: SerdeOptions {
+                            rename: Some(String::from("awesome property"))
+                        },
+                        property_type: String::from("Value")
+                    }]
+                }
+            ]
+        );
+    }
+
     fn add_type(generator: &mut Generator, data_type: DataType, required: bool) -> String {
         let mut definitions = HashMap::new();
 
